###########################################################################
#                                                                         #
# C O P Y R I G H T   N O T I C E                                         #
#  Copyright (c) 2003-5 by:                                               #
#    * California Institute of Technology                                 #
#                                                                         #
#    All Rights Reserved.                                                 #
#                                                                         #
# Permission is hereby granted, free of charge, to any person             #
# obtaining a copy of this software and associated documentation files    #
# (the "Software"), to deal in the Software without restriction,          #
# including without limitation the rights to use, copy, modify, merge,    #
# publish, distribute, sublicense, and/or sell copies of the Software,    #
# and to permit persons to whom the Software is furnished to do so,       #
# subject to the following conditions:                                    #
#                                                                         #
# The above copyright notice and this permission notice shall be          #
# included in all copies or substantial portions of the Software.         #
#                                                                         #
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,         #
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF      #
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND                   #
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS     #
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN      #
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN       #
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE        #
# SOFTWARE.                                                               #
###########################################################################
#
# ymf.py
from cistematic.programs import Program
import os, time, tempfile
from cistematic.core.motif import Motif
if environ.get("CISTEMATIC_TEMP"):
    cisTemp = environ.get("CISTEMATIC_TEMP")
else:
    cisTemp = '/tmp'

class Ymf(Program):
    """ Yeast Motif Finder - from Sinha and Tompa - downloaded from bio.cs.washington.edu
    """
    ymfPath = Program.programRoot + '/ymf3.0/'
    fePath = Program.programRoot + '/explanators1.1.1/src/'
    tempfile.tempdir = cisTemp
    genome = 'yeast'
    seqLength  = '1000'
    motifLength = '8'
    numExplanators = '4'
    explanators = []
    motifs = []
    
    def getSettings(self):
        """ used primarily save settings to an experiment database.
        """
        return (self.genome, self.seqLength, self.motifLength, self.numExplanators)
    
    def setSettings(self, settings):
        """ used primarily retrieve settings from an experiment database.
        """
        self.motifs = []
        (self.genome, self.seqLength, self.motifLength, self.numExplanators) = settings
    
    def setGenExpOptions(self, optionArray):
        for option in optionArray:
            (optionTag, optionValue) = option.split(':')
            if optionTag == 'motifLength':
                self.setLength(optionValue)
        
    def buildCommand(self, tempDir):
        """ returns the command line that will be executed for Ymf.
        """
        cmd = 'cd ' + tempDir + '; ' + self.ymfPath + 'statsvar/statsvar ' + self.ymfPath + 'statsvar/stats.config ' + self.seqLength  + ' ' + self.motifLength + ' ' + self.ymfPath + 'ymftables/' + self.genome + '  ' + self.inputFilePath 
        
        #print "cmd is %s" % (cmd)
        
        return cmd
    
    def buildFEcommand(self, tempDir):
        """ returns the command line that will be executed for findExplanators to process Ymf results.
        """
        cmd = 'cd ' + tempDir + '; ' + self.fePath + 'find_explanators ' + self.ymfPath + 'ymftables/' + self.genome  + '_powersGeneralized.3.bin ' + self.inputFilePath + ' results '  + self.numExplanators       
        #print "cmd is %s" % (cmd)
        
        return cmd

    def setGenome(self, gen):
        """ must specify the genome so that the right stat tables are used.!
        """
        self.genome = gen
    
    def setSeqLength(self, length):
        """ YMF use to only run on fixed length sequences. Just give the longest sequence instead.
        """
        self.seqLength = str(length)

    def setLength(self, length):
        """ set motif length - YMF become exponentially slower at higher lengths - 6 to 10 is the 
          practical range.
        """
        self.motifLength = str(length)
 
    def setExplanators(self, num):
        """ set the number of explanators.
        """
        self.numExplanators = str(num)
    
    # run the program - preferrably after some of the other options have been set
    def run(self):
        """ override Parent's run() method to run YMF and FindExplanators on the
        input file set from Parent.inputFile()
        """
        oldtempdir = tempfile.tempdir
        tempfile.tempdir = Program.programRoot + '/workdir'
        
        ymfTempDir = tempfile.mktemp()
        tempfile.tempdir = oldtempdir
        
        os.mkdir(ymfTempDir)
        startTime = time.time()
        self.contents = os.popen(Ymf.buildCommand(self, ymfTempDir)).readlines()
        self.contents = os.popen(Ymf.buildFEcommand(self, ymfTempDir)).readlines()
        stopTime = time.time()
        
        resFile = open(ymfTempDir + '/results', 'r')
        self.contents = resFile.readlines()
        resFile.close()
        
        expFile = open(ymfTempDir + '/__explanators__', 'r')
        self.explanators = expFile.readlines()
        expFile.close()
        print "\nThis run took %.3f - %.3f = %.3f seconds" % (startTime, stopTime, stopTime - startTime)
    
    def getMotifs(self):
        """ return results from explanators as motifs.
        """
        index = 1
        
        for line in self.explanators:
            self.motifs.append(Motif(self.tagID + '-YMF-' + str(index),line[:-1]))
            index +=1
        
        return self.motifs
